#include "include/common.h"

struct db_t		db_auth;
pthread_mutex_t		db_auth_mutex;

struct auth_t		auth;

int auth_init( char *admin_id, char *admin_pass )
{
	char		*init_sql;

	init_sql = db_printf( CATALOG_INIT_USER_SQL, admin_id, admin_pass );
	if ( init_sql == NULL )
	{
		perror( "Cannot open catalog files[init sql memory allocation failed]" );
		return SQLITE_ERROR;
	}

	if ( db_open_init( &db_auth, "catalog.user.superdb", init_sql ) != SQLITE_OK )
	{
		db_free( init_sql );
		perror( "Cannot open catalog files[catalog.user.superdb]" );
		return SQLITE_ERROR;
	}
	db_free( init_sql );

	pthread_mutex_init( &db_auth_mutex, NULL );

	printf( "** Initializing system catalog: User.		" );
	fflush( stdout );
	if ( auth_user_flush() != SQLITE_OK )
	{
		printf( "[Failed]\n***** Error: %s\n", db_auth.errmsg );
		return SQLITE_ERROR;
	}
	printf( "[OK]\n" );

	printf( "** Initializing system catalog: Grant DB.	" );
	fflush( stdout );
	if ( auth_db_flush() != SQLITE_OK )
	{
		printf( "[Failed]\n***** Error: %s\n", db_auth.errmsg );
		return SQLITE_ERROR;
	}
	printf( "[OK]\n" );

	printf( "** Initializing system catalog: Grant Object.	" );
	fflush( stdout );
	if ( auth_object_flush() != SQLITE_OK )
	{
		printf( "[Failed]\n***** Error: %s\n", db_auth.errmsg );
		return SQLITE_ERROR;
	}
	printf( "[OK]\n" );

	return SQLITE_OK;
}

int auth_user_flush()
{
	int	result;

	result = db_get_table( &db_auth, "SELECT * FROM USER" );
	if ( result != SQLITE_OK )
		return SQLITE_ERROR;

	pthread_mutex_lock( &db_auth_mutex );

	if ( auth.user.table != NULL )
		_db_free_table( auth.user.table );

	auth.user.rows = db_table_num_rows( &db_auth );
	auth.user.cols = db_table_columns( &db_auth );
	auth.user.table = db_auth.table;

	pthread_mutex_unlock( &db_auth_mutex );

	return SQLITE_OK;
}

int auth_db_flush()
{
	int	result;

	result = db_get_table( &db_auth, "SELECT * FROM GRANT_DB" );
	if ( result != SQLITE_OK )
		return SQLITE_ERROR;

	pthread_mutex_lock( &db_auth_mutex );

	if ( auth.db.table != NULL )
		_db_free_table( auth.db.table );

	auth.db.rows = db_table_num_rows( &db_auth );
	auth.db.cols = db_table_columns( &db_auth );
	auth.db.table = db_auth.table;

	pthread_mutex_unlock( &db_auth_mutex );

	return SQLITE_OK;
}

int auth_object_flush()
{
	int	result;

	result = db_get_table( &db_auth, "SELECT * FROM GRANT_OBJECT" );
	if ( result != SQLITE_OK )
		return SQLITE_ERROR;

	pthread_mutex_lock( &db_auth_mutex );

	if ( auth.object.table != NULL )
		_db_free_table( auth.object.table );

	auth.object.rows = db_table_num_rows( &db_auth );
	auth.object.cols = db_table_columns( &db_auth );
	auth.object.table = db_auth.table;

	pthread_mutex_unlock( &db_auth_mutex );

	return SQLITE_OK;
}

#define	AUTH_SELECT		0
#define	AUTH_INSERT		1
#define	AUTH_UPDATE		2
#define	AUTH_DELETE		3
#define	AUTH_ALTER		4
#define	AUTH_TRIGGER		5
#define	AUTH_SHOW		6

#define	AUTH_CREATE_TABLE	0
#define	AUTH_CREATE_VIEW	1
#define	AUTH_CREATE_INDEX	2
#define	AUTH_CREATE_TRIGGER	3
int authorizer( void *data, int auth_code, const char *argv1, const char *argv2, const char *dbversion, const char *inner )
{
	extern short		repl_master_able;

	struct thread_t		*self = (struct thread_t *) data;
	struct db_user_t	*user;
	char			*check;
	short			type = 0;
	int			code = -1;

	const char		*object = NULL;

	if ( data == NULL )
		return SQLITE_DENY;

	user = &self->user;
	if ( user->super == 1 )
		return SQLITE_OK;

	//printf("     %d, %s, %s, %s, %s\n", auth_code, argv1, argv2, dbversion, inner);

	switch ( auth_code )
	{
		/* db manipulation */
		case SQLITE_CREATE_TABLE:		/* Table Name      NULL            */
		case SQLITE_DROP_TABLE:			/* Table Name      NULL            */
		case SQLITE_CREATE_TEMP_TABLE:		/* Table Name      NULL            */
		case SQLITE_DROP_TEMP_TABLE:		/* Table Name      NULL            */
			if ( strcmp( argv1, "sqlite_master" ) == 0 || strcmp( argv1, "sqlite_temp_master" ) == 0 )
				return SQLITE_OK;
			code = AUTH_CREATE_TABLE;
			type = 1;
		break;
		case SQLITE_CREATE_VIEW:		/* View Name       NULL            */
		case SQLITE_DROP_VIEW:			/* View Name       NULL            */
		case SQLITE_CREATE_TEMP_VIEW:		/* View Name       NULL            */
		case SQLITE_DROP_TEMP_VIEW:		/* View Name       NULL            */
			code = AUTH_CREATE_VIEW;
			type = 1;
		break;
		case SQLITE_CREATE_INDEX:		/* Index Name      Table Name      */
		case SQLITE_DROP_INDEX:			/* Index Name      Table Name      */
		case SQLITE_CREATE_TEMP_INDEX:		/* Index Name      Table Name      */
		case SQLITE_DROP_TEMP_INDEX:		/* Index Name      Table Name      */
			code = AUTH_CREATE_INDEX;
			type = 1;
		break;
		case SQLITE_CREATE_TRIGGER:		/* Trigger Name    Table Name      */
		case SQLITE_DROP_TRIGGER:		/* Trigger Name    Table Name      */
		case SQLITE_CREATE_TEMP_TRIGGER:	/* Trigger Name    Table Name      */
		case SQLITE_DROP_TEMP_TRIGGER:		/* Trigger Name    Table Name      */
#ifdef SQLITE_FUNCTION
		case SQLITE_FUNCTION:
#endif
			code = AUTH_CREATE_TRIGGER;
			type = 1;
		break;
		/* access control */
		case SQLITE_TRANSACTION:		/* NULL            NULL            */
		case SQLITE_SELECT:			/* NULL            NULL            */
		case SQLITE_REINDEX:			/* Index Name      NULL            */
		case SQLITE_PRAGMA:			/* Pragma Name     1st arg or NULL */
		case SQLITE_ATTACH:			/* Filename        NULL            */
		case SQLITE_DETACH:			/* Database Name   NULL            */
			return SQLITE_OK;
		break;
#ifdef SQLITE_ANALYZE
		case SQLITE_ANALYZE:			/* Table Name      NULL            */
#endif
		case SQLITE_INSERT:			/* Table Name      NULL            */
			if ( strcmp( argv1, "sqlite_master" ) == 0 || strcmp( argv1, "sqlite_temp_master" ) == 0 )
				return SQLITE_OK;
			object = argv1;
			code = AUTH_INSERT;
			type = 2;
		break;
		case SQLITE_DELETE:			/* Table Name      NULL            */
			if ( strcmp( argv1, "sqlite_master" ) == 0 || strcmp( argv1, "sqlite_temp_master" ) == 0 )
				return SQLITE_OK;
			object = argv1;
			code = AUTH_DELETE;
			type = 2;
		break;
		case SQLITE_READ:			/* Table Name      Column Name     */
			if ( argv1 != NULL && ( strcmp( argv1, "sqlite_master" ) == 0 || strcmp( argv1, "sqlite_temp_master" ) == 0 ) )
				return SQLITE_OK;
			object = argv1;
			code = AUTH_SELECT;
			type = 2;
		break;
		case SQLITE_UPDATE:			/* Table Name      Column Name     */
			if ( strcmp( argv1, "sqlite_master" ) == 0 || strcmp( argv1, "sqlite_temp_master" ) == 0 )
				return SQLITE_OK;
			object = argv1;
			code = AUTH_UPDATE;
			type = 2;
		break;
		case SQLITE_ALTER_TABLE:		/* Database Name   Table Name      */
			if ( strcmp( argv2, "sqlite_master" ) == 0 || strcmp( argv2, "sqlite_temp_master" ) == 0 )
				return SQLITE_OK;
			object = argv2;
			code = AUTH_ALTER;
			type = 2;
		break;
		/* extended */
		case SQLITE_SHOW:
		case SQLITE_DESC:
			object = "*";
			code = AUTH_SHOW;
			type = 2;
		break;
	}

	if ( user->dbname == NULL )
		user->dbname = "*";

	if ( type == 1 )
	{
		check = auth_db_check( user->id, user->dbname, code );
		if ( check == NULL )
			check = auth_db_check( user->id, "*", code );
		if ( check == NULL )
			check = auth_db_check( "*", user->dbname, code );
		if ( check == NULL )
			check = auth_db_check( "*", "*", code );

		if ( check == NULL || check[0] != '1' )
			return SQLITE_DENY;
	}
	else if ( type == 2 )
	{
		check = auth_object_check( user->id, user->dbname, object, code );
		if ( SQLITE_SHOW || SQLITE_DESC )
		{
			if ( check == NULL || check[0] == '1' )
				return SQLITE_OK;
			return SQLITE_DENY;
		}

		if ( check == NULL )
			check = auth_object_check( user->id, user->dbname, "*", code );
		if ( check == NULL )
			check = auth_object_check( user->id, "*", "*", code );

		if ( check == NULL )
			check = auth_object_check( "*", user->dbname, object, code );
		if ( check == NULL )
			check = auth_object_check( "*", user->dbname, "*", code );
		if ( check == NULL )
			check = auth_object_check( "*", "*", "*", code );

		if ( check == NULL )
			check = auth_object_check( user->id, "*", object, code );
		if ( check == NULL )
			check = auth_object_check( "*", "*", object, code );

		if ( check == NULL || check[0] != '1' )
			return SQLITE_DENY;
	}

	if ( repl_master_able == TRUE )
	{
		if ( type == 1 || (type == 2 && code == SQLITE_ALTER_TABLE ) )
			repl_master_updated( self );
	}

	return SQLITE_OK;
}

#define	AUTH_CREATE_ALL		4
#define	AUTH_ALL		7

int auth_update( short type, struct grant_t *data )
{
	char	*curval;
	char	tmp[AUTH_ALL+1];
	short	rule = -1;
	char	*sql;
	int	result;

	if ( strncasecmp( data->rule, "all", 3 ) == 0 )
		rule = AUTH_ALL;
	else if ( strncasecmp( data->rule, "select", 6 ) == 0 )
		rule = AUTH_SELECT;
	else if ( strncasecmp( data->rule, "insert", 6 ) == 0 )
		rule = AUTH_INSERT;
	else if ( strncasecmp( data->rule, "update", 6 ) == 0 )
		rule = AUTH_UPDATE;
	else if ( strncasecmp( data->rule, "delete", 6 ) == 0 )
		rule = AUTH_DELETE;
	else if ( strncasecmp( data->rule, "alter", 5 ) == 0 )
		rule = AUTH_ALTER;
	else if ( strncasecmp( data->rule, "trigger", 7 ) == 0 )
		rule = AUTH_TRIGGER;
	else if ( strncasecmp( data->rule, "show", 4 ) == 0 )
		rule = AUTH_SHOW;

	if ( rule < 0 )
		return SQLITE_ERROR;

	if ( data->object == NULL )
		data->object = "*";

	if ( strcasecmp( data->user, "public" ) == 0 )
		data->user = "*";

	curval = auth_object_check( data->user, data->dbname, data->object, AUTH_ALL );
	if ( curval == NULL )
	{
		memset( tmp, '0', AUTH_ALL );
		curval = tmp;
	}

	if ( rule == AUTH_ALL )
		curval[0] = curval[1] = curval[2] = curval[3] = curval[4] = curval[5] = curval[6] = curval[7] = (type == SQLITE_AUTH_GRANT?'1':'0');
	else
		curval[rule] = (type == SQLITE_AUTH_GRANT?'1':'0');

	curval[AUTH_ALL] = '\0';

	sql = db_printf( "REPLACE INTO GRANT_OBJECT (id, db, object, grant) VALUES('%q', '%q', '%q', '%q')", data->user, data->dbname, data->object, curval );
	result = db_throw( &db_auth, sql );
	db_free( sql );

	auth_object_flush();

	return result;
}

int auth_create_update( short type, struct grant_t *data )
{
	char	*curval;
	char	tmp[AUTH_CREATE_ALL+1];
	short	rule = -1;
	char	*sql;
	int	result;

	if ( strncasecmp( data->rule, "all", 3 ) == 0 )
		rule = AUTH_CREATE_ALL;
	else if ( strncasecmp( data->rule, "table", 5 ) == 0 )
		rule = AUTH_CREATE_TABLE;
	else if ( strncasecmp( data->rule, "view", 4 ) == 0 )
		rule = AUTH_CREATE_VIEW;
	else if ( strncasecmp( data->rule, "index", 5 ) == 0 )
		rule = AUTH_CREATE_INDEX;
	else if ( strncasecmp( data->rule, "trigger", 7 ) == 0 )
		rule = AUTH_CREATE_TRIGGER;

	if ( rule < 0 )
		return SQLITE_ERROR;

	if ( strcasecmp( data->user, "public" ) == 0 )
		data->user = "*";

	curval = auth_db_check( data->user, data->dbname, AUTH_CREATE_ALL );
	if ( curval == NULL )
	{
		memset( tmp, '0', AUTH_CREATE_ALL );
		curval = tmp;
	}

	if ( rule == AUTH_CREATE_ALL )
		curval[0] = curval[1] = curval[2] = curval[3] = (type == SQLITE_AUTH_GRANT_CREATE?'1':'0');
	else
		curval[rule] = (type == SQLITE_AUTH_GRANT_CREATE?'1':'0');

	curval[AUTH_CREATE_ALL] = '\0';

	sql = db_printf( "REPLACE INTO GRANT_DB (id, db, grant) VALUES('%q', '%q', '%q')", data->user, data->dbname, curval );
	result = db_throw( &db_auth, sql );
	db_free( sql );

	auth_db_flush();

	return result;
}

#undef	AUTH_SELECT
#undef	AUTH_INSERT
#undef	AUTH_UPDATE
#undef	AUTH_DELETE
#undef	AUTH_ALTER
#undef	AUTH_TRIGGER
#undef	AUTH_SHOW

#undef	AUTH_CREATE_TABLE
#undef	AUTH_CREATE_VIEW
#undef	AUTH_CREATE_INDEX
#undef	AUTH_CREATE_TRIGGER

int auth_parse_str( char *id, char **pass, char *auth_str )
{
	char	iddata[1024];
	int	iddata_len;

	iddata_len = base64_decode( iddata, auth_str, strlen(auth_str) );

	if ( iddata_len <= 0 )
		return SQLITE_AUTH;
	iddata[iddata_len] = '\0';

	memcpy( id, iddata, iddata_len+1 );
	(*pass) = strchr( id, ':' );
	if ( (*pass) == NULL )
		return SQLITE_AUTH;

	(*pass)[0] = '\0';
	(*pass)++;

	return SQLITE_OK;
}

#define	AUTH_FIELD_ID		0
#define	AUTH_FIELD_PASS		1
#define	AUTH_FIELD_SUPER	2
int auth_user_check( char *id, char *pass, struct db_user_t *result )
{
	int	i;
	for ( i=0; i < auth.user.rows; i++ )
	{
		if ( strcasecmp( auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_ID)], id ) == 0 )
		{
			if ( strcasecmp( auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_PASS)], pass ) == 0 )
			{
				snprintf( result->id, 128, "%s", auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_ID)] );
				snprintf( result->pass , 128, "%s", auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_PASS)] );
				result->super = atoi( auth.user.table[(auth.user.cols * (i+1) + AUTH_FIELD_SUPER)] );
				return SQLITE_OK;
			}
			return SQLITE_AUTH;
		}
	}

	return SQLITE_AUTH;
}
#undef	AUTH_FIELD_PASS
#undef	AUTH_FIELD_SUPER


#define	AUTH_FIELD_DB		1
#define	AUTH_FIELD_GRANT	2
char *auth_db_check( const char *id, const char *db, int grant_type )
{
	int	i;

#ifdef __DEBUG
	printf("----- DB: %s, %s, %d\n", id, db, grant_type);
#endif
	for ( i=0; i < auth.db.rows; i++ )
	{
		if ( strcasecmp( auth.db.table[(auth.db.cols * (i+1) + AUTH_FIELD_ID)], id ) == 0 && strcasecmp( auth.db.table[(auth.db.cols * (i+1) + AUTH_FIELD_DB)], db ) == 0 )
		{
			if ( grant_type == AUTH_CREATE_ALL )
				return auth.db.table[(auth.db.cols * (i+1) + AUTH_FIELD_GRANT)];
			else
				return &(auth.db.table[(auth.db.cols * (i+1) + AUTH_FIELD_GRANT)][grant_type]);
		}
	}

	return NULL;
}
#undef	AUTH_FIELD_GRANT


#define	AUTH_FIELD_OBJECT	2
#define	AUTH_FIELD_GRANT	3
char *auth_object_check( const char *id, const char *db, const char *object, int grant_type )
{
	int	i;

#ifdef __DEBUG
	printf("----- OBJECT: %s, %s, %s, %d\n", id, db, object, grant_type);
#endif
	for ( i=0; i < auth.object.rows; i++ )
	{
		if ( strcasecmp( auth.object.table[(auth.object.cols * (i+1) + AUTH_FIELD_ID)], id ) == 0 && strcasecmp( auth.object.table[(auth.object.cols * (i+1) + AUTH_FIELD_DB)], db ) == 0 && strcasecmp( auth.object.table[(auth.object.cols * (i+1) + AUTH_FIELD_OBJECT)], object ) == 0 )
		{
			if ( grant_type == AUTH_ALL )
				return auth.object.table[(auth.object.cols * (i+1) + AUTH_FIELD_GRANT)];
			else
				return &(auth.object.table[(auth.object.cols * (i+1) + AUTH_FIELD_GRANT)][grant_type]);
		}
	}

	return NULL;
}
#undef	AUTH_FIELD_ID
#undef	AUTH_FIELD_OBJECT
#undef	AUTH_FIELD_GRANT
