#include "include/common.h"

struct cache_t	*cache = NULL;
short		cache_able = FALSE;
int		cache_slot = 0;
int		cache_buf_size = 0;
pthread_mutex_t	cache_lock;

int cache_init( int cache_cnt, int buf_size )
{
	size_t	size;
	int	i;

	size = sizeof( struct cache_t ) * cache_cnt;
	cache = malloc( size );
	if ( cache == NULL )
		return SQLITE_ERROR;

	memset( cache, 0, size );
	for ( i=0; i < cache_cnt; i++ )
		cache[i].buf = malloc( buf_size+1 );

	pthread_mutex_init( &cache_lock, NULL );

	cache_slot = cache_cnt;
	cache_buf_size = buf_size;

#ifdef	SQLITE_ENABLE_UPDATE_HOOK
	cache_able = TRUE;
#endif

	return SQLITE_OK;
}

void cache_release_hook( struct db_t *db )
{
	//DEBUG2("DB is updated")
	if ( db != NULL )
		db->updated = TRUE;
	return;
}

void cache_release( char *db_name )
{
	size_t		db_name_len;
	int		i;

	if ( cache_slot <= 0 || db_name == NULL )
		return;

	db_name_len = strlen(db_name);

	pthread_mutex_lock( &cache_lock );
	for ( i=0; i < cache_slot; i++ )
	{
		if ( cache[i].hit > 0 && strncmp( db_name, cache[i].db_name, db_name_len ) == 0 )
		{
			//DEBUG2("cache release")
			cache[i].hit = 0;
		}
	}
	pthread_mutex_unlock( &cache_lock );
	return;
}

void cache_push( char *db_name, char *query, long long int rowid, int rows, char *buf, int size )
{
	struct cache_t	*node = NULL;
	size_t		sql_len;
	size_t		db_name_len;

	if ( size > cache_buf_size || cache_slot <= 0 || query == NULL || buf == NULL || size <= 0 )
		return;

	sql_len = strlen( query );
	db_name_len = strlen(db_name);

	pthread_mutex_lock( &cache_lock );
	node = cache_get_old();
	if ( node != NULL )
	{
		//DEBUG2("cache pushing")
		if ( node->sql_buf_size <= sql_len )
		{
			if ( node->sql != NULL )
				free( node->sql );
			node->sql_buf_size = sql_len;
			node->sql = malloc( sql_len+1 );
		}
		
		memcpy( node->sql, query, sql_len );
		node->sql[sql_len] = '\0';
		node->sql_len = sql_len;
		
		memcpy( node->buf, buf, size );
		node->buf[size] = '\0';
		node->buf_len = size;
		
		memcpy( node->db_name, db_name, db_name_len );
		node->db_name[db_name_len] = '\0';
		
		gettimeofday( &(node->last), NULL );
		
		node->hit = 1;
		node->rowid = rowid;
		node->rows = rows;
	}
	pthread_mutex_unlock( &cache_lock );
	return;
}

int cache_get( char *db_name, struct cache_t *qcache )
{
	size_t		sql_len;
	size_t		db_name_len;
	int		i;

	if ( cache_slot <= 0 || db_name == NULL )
		return SQLITE_ERROR;

	sql_len = strlen( qcache->sql );
	db_name_len = strlen(db_name);

	pthread_mutex_lock( &cache_lock );
	for ( i=0; i < cache_slot; i++ )
	{
		if ( cache[i].hit <= 0 || cache[i].sql_len != sql_len )
			continue;
		if ( strncmp( db_name, cache[i].db_name, db_name_len ) == 0 && strncmp( qcache->sql, cache[i].sql, sql_len ) == 0 )
		{
			//DEBUG2("cache found")
			qcache->rowid = cache[i].rowid;
			qcache->rows = cache[i].rows;
			qcache->buf = cache[i].buf;
			qcache->buf_len = cache[i].buf_len;
			cache[i].hit++;
			gettimeofday( &(cache[i].last), NULL );
			pthread_mutex_unlock( &cache_lock );
			return SQLITE_OK;
		}
	}
	pthread_mutex_unlock( &cache_lock );
	return SQLITE_ERROR;
}

struct cache_t *cache_get_old()
{
	int	i;
	int	idx = 0;
	int	time_sec;
	int	time_usec;

	time_sec = cache[0].last.tv_sec;
	time_usec = cache[0].last.tv_usec;

	for ( i=1; i < cache_slot; i++ )
	{
		if ( cache[i].hit <= 0 )
			return &(cache[i]);
		if ( time_sec > cache[i].last.tv_sec || ( time_sec == cache[i].last.tv_sec && time_usec > cache[i].last.tv_usec ) )
		{
			idx = i;
			time_sec = cache[i].last.tv_sec;
			time_usec = cache[i].last.tv_usec;
		}
	}
	return &(cache[idx]);
}

void cache_safe_exit()
{
	int	i;
	
	if ( cache_slot <= 0 )
		return;

	for ( i=0; i < cache_slot; i++ )
	{
		if ( cache[i].buf != NULL )
			free( cache[i].buf );
		if ( cache[i].sql != NULL )
			free( cache[i].sql );
	}
	pthread_mutex_destroy( &cache_lock );
	cache_able = FALSE;
	return;
}
