#include "include/common.h"

static int		db_work_id = 0;

struct db_lock_t	*db_lock_pool;
pthread_mutex_t		db_lock_pool_lock;
int			db_lock_pool_cnt = 0;

void db_init()
{
	pthread_mutex_init( &db_lock_pool_lock, NULL );
}

struct db_lock_t *db_lock_pool_get( char *db_name )
{
	int	i;

	if ( db_lock_pool_cnt > 0 )
	{
		for ( i=0; i < db_lock_pool_cnt; i++ )
		{
			if ( strcasecmp( db_lock_pool[i].db_name, db_name ) == 0 )
				return &(db_lock_pool[i]);
		}
	}

	pthread_mutex_lock( &db_lock_pool_lock );
	db_lock_pool = realloc( db_lock_pool, sizeof(struct db_lock_t)*(db_lock_pool_cnt+1) );
	memset( &(db_lock_pool[db_lock_pool_cnt]), 0, sizeof(struct db_lock_t) );
	snprintf( db_lock_pool[db_lock_pool_cnt].db_name, LIMIT_MAX_DBNAME, "%s", db_name );
	pthread_mutex_init( &(db_lock_pool[db_lock_pool_cnt].lock), NULL );
	i = db_lock_pool_cnt;
	db_lock_pool_cnt++;
	pthread_mutex_unlock( &db_lock_pool_lock );

	return &(db_lock_pool[i]);
}

int db_open( struct db_t *db, char *filename )
{
	extern struct config_t	config;
	extern int		errno;
	
	struct stat		stat;
	int			db_res;
	char			filepath[1024*4];

	if ( db == NULL )
		return SQLITE_ERROR;

	if ( strncmp( DB_MEMORY, filename, DB_MEMORY_LEN ) == 0 )
	{
		db_res = sqlite3_open( DB_MEMORY, &db->db );
		if ( db_res == SQLITE_OK )
		{
			db->opened = TRUE;
			gettimeofday( &(db->open_time), NULL );
			return SQLITE_OK;
		}
	}
	else
	{
		snprintf( filepath, 1024*4, "%s/%s", config.db_path, filename );
		if ( lstat( filepath, &stat ) == 0 )
		{
			if ( S_ISREG(stat.st_mode) || S_ISLNK(stat.st_mode) )
			{
				db_res = sqlite3_open( filepath, &db->db );
				if ( db_res == SQLITE_OK )
				{
					db->opened = TRUE;
					gettimeofday( &(db->open_time), NULL );
					sqlite3_busy_handler( db->db, db_busy_callback, NULL );
					return SQLITE_OK;
				}
				return db_res;
			}
		}
	}
	return SQLITE_NOTFOUND;
}

int db_open_init( struct db_t *db, char *filename, char *init_sql )
{
	extern struct config_t	config;
	
	struct stat		stat;
	int			db_res = SQLITE_ERROR;
	short			need_init = TRUE;
	char			filepath[1024*4];

	if ( db == NULL )
		return SQLITE_ERROR;

	if ( strncmp( DB_MEMORY, filename, DB_MEMORY_LEN ) == 0 )
	{
		db_res = sqlite3_open( DB_MEMORY, &db->db );

		if ( db_res == SQLITE_OK )
		{
			db->opened = TRUE;
			gettimeofday( &(db->open_time), NULL );
			if ( need_init == TRUE && init_sql != NULL )
			{
				db_throw( db, init_sql );
				db_err_free( db );
			}
			sqlite3_busy_handler( db->db, db_busy_callback, NULL );
			return SQLITE_OK;
		}
	}
	else
	{
		snprintf( filepath, 1024*4, "%s/%s", config.db_path, filename );
		if ( lstat( filepath, &stat ) == 0 && ( S_ISREG(stat.st_mode) || S_ISLNK(stat.st_mode) ) )
			need_init = FALSE;

		db_res = sqlite3_open( filepath, &db->db );
		if ( db_res == SQLITE_OK )
		{
			db->opened = TRUE;
			gettimeofday( &(db->open_time), NULL );
			if ( need_init == TRUE && init_sql != NULL )
			{
				db_throw( db, init_sql );
				db_err_free( db );
			}
			return SQLITE_OK;
		}
	}
	return db_res;
}

int db_close( struct db_t *db )
{
	int	result;

	if ( db->opened == TRUE )
	{
		db_err_free( db );
		db_commit( db );
		result = sqlite3_close( db->db );
		db->opened = FALSE;
		db->db = NULL;
		return result;
	}
	return SQLITE_OK;
}

int db_set_authorizer( struct db_t *db, void *data )
{
	return sqlite3_set_authorizer( db->db, authorizer, data );
}

int db_set_replicate( struct db_t *db, void *data )
{
	return sqlite3_set_authorizer( db->db, repl_master_authorizer, data );
}

int db_throw( struct db_t *db, char *sql )
{
	if ( db == NULL || db->db == NULL )
		return SQLITE_ERROR;
	db->work_id = ++db_work_id;
	return sqlite3_exec( db->db, sql, NULL, NULL, NULL );
}

int db_commit( struct db_t *db )
{
	return db_throw( db, "COMMIT" );
}

int db_rollback( struct db_t *db )
{
	return db_throw( db, "ROLLBACK" );
}

char *db_printf( const char *sql, ...)
{
	va_list	argv;
	char	*dst;

	va_start( argv, sql );
	dst = sqlite3_vmprintf( sql, argv );
	va_end( argv );

	return dst;
}

void db_err_free( struct db_t *db )
{
	if ( db->errmsg != NULL )
	{
		db_free( db->errmsg );
		db->errmsg = NULL;
	}
}

void db_free( char *sql )
{
	sqlite3_free( sql );
}

int db_changes( struct db_t *db )
{
	if ( db->db == NULL )
		return 0;
	return sqlite3_changes( db->db );
}

long long int db_last_rowid( struct db_t *db )
{
	if ( db->db == NULL )
		return 0;
	return sqlite3_last_insert_rowid( db->db );
}

int db_get_table( struct db_t *db, char *sql )
{
	int	res;

	if ( db == NULL )
		return SQLITE_ERROR;
	db->work_id = ++db_work_id;

	if ( db->lock != NULL )
		pthread_mutex_lock( &db->lock->lock );

	res = sqlite3_get_table( db->db, sql, &db->table, &db->table_nrow, &db->table_ncol, &db->errmsg );

	if ( db->lock != NULL )
		pthread_mutex_unlock( &db->lock->lock );

	return res;
}

int db_busy_callback( void *data, int count )
{
	if ( count >= 10 )
		return 0;
	usleep( 10000 );
	return SQLITE_ERROR;
}

int db_table_columns( struct db_t *db )
{
	if ( db->table_res == SQLITE_OK )
		return db->table_ncol;
	return -1;
}

int db_table_column_name( char **column_name, struct db_t *db, int offset )
{
	if ( db->table_res == SQLITE_OK )
	{
		(*column_name) = db->table[offset];
		return strlen( (*column_name) );
	}
	else
	{
		(*column_name) = "";
		return 0;
	}
}

int db_table_num_rows( struct db_t *db )
{
	if ( db->table_res == SQLITE_OK )
		return db->table_nrow;
	return -1;
}

int db_table_has_rows( struct db_t *db )
{
	if ( db->table_res == SQLITE_OK )
		return (db->table_nrow > 0)?SQLITE_OK:SQLITE_ERROR;
	return SQLITE_ERROR;
}

char *db_table_result( struct db_t *db, int row, int col )
{
	if ( db->table_res == SQLITE_OK && db->table_nrow >= row )
		return db->table[(db->table_ncol * (row+1) + col)];
	return NULL;
}

void db_free_table( struct db_t *db )
{
	if ( db->table != NULL )
	{
		_db_free_table( db->table );
		db->table = NULL;
	}

	db->table_res = db->table_nrow = db->table_ncol = -1;
}

void _db_free_table( char **table )
{
	return sqlite3_free_table( table );
}

const char *db_version()
{
	return sqlite3_libversion();
}

int db_column_meta( struct db_t *db, char *table, const char *column, struct desc_t *desc )
{
#ifdef SQLITE_ENABLE_COLUMN_METADATA
	if ( db == NULL )
		return SQLITE_ERROR;
	return sqlite3_table_column_metadata( db->db, NULL, table, column, &(desc->type), &(desc->seq), &(desc->notnull), &(desc->pkey), &(desc->autoincrement) );
#else
	return SQLITE_ERROR;
#endif
}

const char *db_error( struct db_t *db )
{
	if ( db == NULL || db->db == NULL )
		return "No database";
	return sqlite3_errmsg( db->db );
}

int db_statement( struct db_t *db, char *sql )
{
	if ( db == NULL )
		return SQLITE_ERROR;
	db->work_id = ++db_work_id;
	return sqlite3_prepare( db->db, sql, strlen(sql), &(db->stmt), db->sql_tail );
}

int db_statement_column_count( struct db_t *db )
{
	if ( db == NULL || db->stmt == NULL )
		return SQLITE_ERROR;
	return sqlite3_column_count( db->stmt );
}

const char *db_statement_column_type( struct db_t *db, int seq )
{
	if ( db == NULL || db->stmt == NULL )
		return "";
	return sqlite3_column_decltype( db->stmt, seq );
}

const char *db_statement_column_name( struct db_t *db, int seq )
{
	if ( db == NULL || db->stmt == NULL )
		return "";
	return sqlite3_column_name( db->stmt, seq );
}

int db_statement_free( struct db_t *db )
{
	int	result;
	result=sqlite3_finalize( db->stmt );
	db->stmt = NULL;
	return result;
}

void db_thread_cleanup()
{
	sqlite3_thread_cleanup();
}

