#include "include/common.h"

int pstmt_process( struct thread_t *self )
{
	unsigned long	pstmt_id;
	char		*extend;
	int		bind_seq;
	char		*bind_type;

	if ( self->request.header.extend == NULL )
		return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );

	pstmt_id = strtoul( self->request.header.extend, &extend, 10 );
	if ( extend == NULL )
		return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
	extend++;

	switch (extend[0])
	{
		case 'i':
		case 'I':
			//init
			if ( strncasecmp( &(extend[1]), "NIT", 3 ) == 0 )
			{
				return pstmt_init( self, pstmt_id, self->request.data );
			}
			else
				return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
		break;
		case 'r':
		case 'R':
			//release, reset
			if ( strncasecmp( &(extend[1]), "ELEASE", 6 ) == 0 )
			{
				return pstmt_release( self, pstmt_id );
			}
			else if ( strncasecmp( &(extend[1]), "ESET", 4 ) == 0 )
			{
				return pstmt_reset( self, pstmt_id );
			}
			else
				return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
		break;
		case 'e':
		case 'E':
			//execute
			if ( strncasecmp( &(extend[1]), "XECUTE", 6 ) == 0 )
			{
				return pstmt_execute( self, pstmt_id );
			}
			else
				return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
		break;
		default:
			//bind
			bind_seq = atoi( extend );
			bind_type = strchr( extend, '/' );
			if ( bind_type == NULL )
				return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
			bind_type++;
			return pstmt_bind( self, pstmt_id, bind_seq, bind_type, self->request.data, self->request.data_len );
		break;
	}

	return SQLITE_OK;
}

int pstmt_init( struct thread_t *self, unsigned long pstmt_id, char *sql )
{
	int		result;
	struct pstmt_t	*pool;
	struct pstmt_t	*pstmt;
	sqlite3_stmt	*stmt;
	int		param_cnt;
	size_t		sql_len;


	self->db.pstmt_run = -1;

	sql_len = strlen( sql );

	result = sqlite3_prepare( self->db.db, sql, sql_len, &stmt, self->db.sql_tail );
	if ( result != SQLITE_OK )
		return response_error( self, RESPONSE_ERR_INTERNAL, (char *) db_error( &self->db ) );

	param_cnt = sqlite3_bind_parameter_count( stmt );

	pstmt_internal_release( self, pstmt_id );
	pool = malloc( sizeof(struct pstmt_t) * (self->db.pstmt_cnt+1) );
	if ( pool == NULL )
	{
		sqlite3_finalize( stmt );
		return response_error( self, RESPONSE_ERR_INTERNAL, NULL );
	}

	memcpy( pool, self->db.pstmt, sizeof(struct pstmt_t) * self->db.pstmt_cnt );
	pstmt = &(pool[self->db.pstmt_cnt]);
	memset( pstmt, 0, sizeof(struct pstmt_t) );

	pstmt->id = pstmt_id;
	pstmt->stmt = stmt;
	pstmt->param_cnt = param_cnt;

	pstmt->sql = malloc( sql_len + 1 );
	strcpy( pstmt->sql, sql );

	pstmt->nolog = TRUE;

	self->db.pstmt = pool;
	self->db.pstmt_cnt++;

	return response_message( self, "OK" );
}

int pstmt_release( struct thread_t *self, unsigned long pstmt_id )
{
	self->db.pstmt_run = -1;
	switch ( pstmt_internal_release( self, pstmt_id ) )
	{
		case SQLITE_OK:
			return response_message( self, "OK" );
		break;
		case SQLITE_NOTFOUND:
		case SQLITE_ERROR_PSTMT_EMPTY:
			return response_error( self, RESPONSE_ERR_DB_FOUND, "Prepared statement not found." );
		break;
		default:
			return response_error( self, RESPONSE_ERR_INTERNAL, NULL );
		break;
	}
	return SQLITE_OK;
}

int pstmt_reset( struct thread_t *self, unsigned long pstmt_id )
{
#ifndef SQLITE_ENABLE_CLEAR_BINDINGS
	int		seq;
#endif
	struct pstmt_t	*pstmt;

	self->db.pstmt_run = -1;
	pstmt = pstmt_internal_find( self, pstmt_id, NULL );
	if ( pstmt == NULL )
		return response_error( self, RESPONSE_ERR_DB_FOUND, "Prepared statement not found." );

	sqlite3_reset( pstmt->stmt );
#ifdef SQLITE_ENABLE_CLEAR_BINDINGS
	if ( sqlite3_clear_bindings( pstmt->stmt ) != SQLITE_OK )
		db_err_free( &(self->db) );
#else
	if ( pstmt->param_cnt > 0 )
	{
		for ( seq = pstmt->param_cnt; seq--; )
		{
			if ( sqlite3_bind_null( pstmt->stmt, seq ) != SQLITE_OK )
				db_err_free( &(self->db) );
		}
	}
#endif

	return response_message( self, "OK" );
}

int pstmt_bind( struct thread_t *self, unsigned long pstmt_id, int bind_seq, char *bind_type, char *data, int data_len )
{
	int			result;
	struct pstmt_t		*pstmt;

	long long int		bind_int;
	double			bind_double;

	self->db.pstmt_run = -1;
	if ( bind_seq < 0 || bind_type == NULL || strlen(bind_type) <= 0 )
		return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );

	pstmt = pstmt_internal_find( self, pstmt_id, NULL );
	if ( pstmt == NULL )
		return response_error( self, RESPONSE_ERR_DB_FOUND, "Prepared statement not found." );

	if ( pstmt->params == NULL )
	{
		pstmt->params = malloc( sizeof(struct pstmt_param_t) );
		pstmt->params_cnt = 0;
	}
	else
		pstmt->params = realloc( pstmt->params, sizeof(struct pstmt_param_t) * (pstmt->params_cnt+1) );

#define PSTMT_PARAM	pstmt->params[pstmt->params_cnt]
	switch ( bind_type[0] )
	{
		case 'i':
		case 'I':
			//integer, integer64
			bind_int = strtoll( data, NULL, 10 );
			result = sqlite3_bind_int64( pstmt->stmt, bind_seq, bind_int );
			if ( result == SQLITE_OK )
			{
				PSTMT_PARAM.type = 'I';
				PSTMT_PARAM.idx = bind_seq;
				PSTMT_PARAM.data.value_integer = bind_int;
			}
		break;
		case 't':
		case 'T':
			//text
			result = sqlite3_bind_text( pstmt->stmt, bind_seq, data, data_len, SQLITE_TRANSIENT );
			if ( result == SQLITE_OK )
			{
				PSTMT_PARAM.type = 'T';
				PSTMT_PARAM.idx = bind_seq;
				PSTMT_PARAM.data_len = data_len;
				PSTMT_PARAM.data.value_text = malloc( data_len+1 );
				memcpy( PSTMT_PARAM.data.value_text, data, data_len );
			}
		break;
		case 'n':
		case 'N':
			//null
			result = sqlite3_bind_null( pstmt->stmt, bind_seq );
			if ( result == SQLITE_OK )
			{
				PSTMT_PARAM.type = 'N';
				PSTMT_PARAM.idx = bind_seq;
			}
		break;
		case 'd':
		case 'D':
			//double
			bind_double = strtod( data, NULL );
			result = sqlite3_bind_double( pstmt->stmt, bind_seq, bind_double );
			if ( result == SQLITE_OK )
			{
				PSTMT_PARAM.type = 'D';
				PSTMT_PARAM.idx = bind_seq;
				PSTMT_PARAM.data.value_double = bind_double;
			}
		break;
		case 'b':
		case 'B':
			//blob
			result = sqlite3_bind_blob( pstmt->stmt, bind_seq, data, data_len, SQLITE_TRANSIENT );
			if ( result == SQLITE_OK )
			{
				PSTMT_PARAM.type = 'B';
				PSTMT_PARAM.idx = bind_seq;
				PSTMT_PARAM.data_len = data_len;
				PSTMT_PARAM.data.value_blob = malloc( data_len+1 );
				memcpy( PSTMT_PARAM.data.value_blob, data, data_len );
			}
		break;
		default:
			return response_error( self, RESPONSE_ERR_BAD_REQUEST, NULL );
		break;
	}
#undef PSTMT_PARAM

	pstmt->params_cnt++;
	pstmt->nolog = TRUE;

	if ( result == SQLITE_OK )
		return response_message( self, "OK" );
	result = response_error( self, RESPONSE_ERR_INTERNAL, (char *) db_error( &self->db ) );
	db_err_free( &(self->db) );
	return result;
}

#define R_BUF		&(self->response.buf[self->response.offset])
#define R_BUF_LIMIT	(self->response.buf_size-self->response.offset)
int pstmt_execute( struct thread_t *self, unsigned long pstmt_id )
{
	int		result;
	struct pstmt_t	*pstmt;

	int		change;
	int		idx = -1;
	long long int	rowid;

	pstmt = pstmt_internal_find( self, pstmt_id, &idx );
	if ( pstmt == NULL )
		return response_error( self, RESPONSE_ERR_DB_FOUND, "Prepared statement not found." );

	self->db.pstmt_run = idx;
	pstmt->nolog = FALSE;

	while ( (result = sqlite3_step( pstmt->stmt )) == SQLITE_BUSY )
	{
		usleep( 200 );
		db_err_free( &(self->db) );
	}

	switch ( result )
	{
		case SQLITE_DONE:
			//no data, updated
			change = db_changes( &self->db );

			//reset cache
			if ( change > 0 )
				cache_release( self->db.db_name );
			rowid = db_last_rowid( &self->db );

			self->response.start = self->response.offset = 0;
			self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, change, rowid, 0, &self->bench_start );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );

			self->response.status = 200;
			self->response.status_msg = "OK";

			sqlite3_reset( pstmt->stmt );

			return response( self );
		break;
		case SQLITE_ROW:
			result = pstmt_internal_row( self, pstmt );
			sqlite3_reset( pstmt->stmt );
			pstmt_internal_gc( self );
			if ( result == SQLITE_OK )
				return SQLITE_OK;
			else
				return SQLITE_ERROR;
		break;
	}

	result = response_error( self, RESPONSE_ERR_INTERNAL, (char *) db_error( &self->db ) );
	db_err_free( &(self->db) );
	return result;
}

int pstmt_internal_row( struct thread_t *self, struct pstmt_t *pstmt )
{
	int		result;

	int		change;
	long long int	rowid;
	int		rows = 0;

	int		column_cnt;
	int		*column_types;

	char		*column_name;
	int		column_name_len;

	int		data_len;

	int		i, p;

	change = db_changes( &self->db );
	rowid = db_last_rowid( &self->db );
	column_cnt = sqlite3_column_count( pstmt->stmt );

	if ( self->db.pstmt_buf == NULL )
	{
		self->db.pstmt_buf_size = 1024 * 1024 * 8;
		self->db.pstmt_buf = malloc( self->db.pstmt_buf_size );
	}
	self->db.pstmt_buf_len = 0;

	if ( self->db.pstmt_tree == NULL )
	{
		self->db.pstmt_tree_size = 1024 * 8;
		self->db.pstmt_tree = malloc( sizeof(char **) * self->db.pstmt_tree_size );
		memset( self->db.pstmt_tree, 0, sizeof(char **) * self->db.pstmt_tree_size );
	}

	column_types = malloc( sizeof(int) * column_cnt );
	self->db.pstmt_tree[rows] = malloc( sizeof(char *) * column_cnt );
	for ( i=0; i < column_cnt; i++ )
	{
		column_types[i] = sqlite3_column_type( pstmt->stmt, i );
		data_len = sqlite3_column_bytes( pstmt->stmt, i );
		if ( column_types[i] == SQLITE_BLOB )
			memcpy( &(self->db.pstmt_buf[self->db.pstmt_buf_len]), sqlite3_column_blob( pstmt->stmt, i ), data_len );
		else
			memcpy( &(self->db.pstmt_buf[self->db.pstmt_buf_len]), sqlite3_column_text( pstmt->stmt, i ), data_len );
		self->db.pstmt_buf[self->db.pstmt_buf_len+data_len] = '\0';
		self->db.pstmt_tree[rows][i] = &(self->db.pstmt_buf[self->db.pstmt_buf_len]);
		self->db.pstmt_buf_len += data_len+1;
	}
	rows++;

	while ( (result = sqlite3_step( pstmt->stmt )) == SQLITE_ROW )
	{
		if ( self->db.pstmt_buf_len > (self->db.pstmt_buf_size + (1024 * 8)) )
		{
			self->db.pstmt_buf_size += 1024 * 1024;
			self->db.pstmt_buf = realloc( self->db.pstmt_buf, self->db.pstmt_buf_size );
		}

		if ( rows > (self->db.pstmt_tree_size + 10) )
		{
			self->db.pstmt_tree_size += 1024 * 8;
			self->db.pstmt_tree = realloc( self->db.pstmt_tree, sizeof(char **) * self->db.pstmt_tree_size );
			memset( &(self->db.pstmt_tree[rows]), 0, sizeof(char **) * (self->db.pstmt_tree_size-rows) );
		}

		self->db.pstmt_tree[rows] = malloc( sizeof(char *) * column_cnt );
		for ( i=0; i < column_cnt; i++ )
		{
			data_len = sqlite3_column_bytes( pstmt->stmt, i );
			if ( column_types[i] == SQLITE_BLOB )
				memcpy( &(self->db.pstmt_buf[self->db.pstmt_buf_len]), sqlite3_column_blob( pstmt->stmt, i ), data_len );
			else
				memcpy( &(self->db.pstmt_buf[self->db.pstmt_buf_len]), sqlite3_column_text( pstmt->stmt, i ), data_len );
			self->db.pstmt_buf[self->db.pstmt_buf_len+data_len] = '\0';
			self->db.pstmt_tree[rows][i] = &(self->db.pstmt_buf[self->db.pstmt_buf_len]);
			self->db.pstmt_buf_len += data_len+1;
		}
		rows++;
	}

	self->response.start = self->response.offset = 0;
	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, change, rowid, rows, &self->bench_start );

	self->response.status = 200;
	self->response.status_msg = "OK";

	//columns open
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
	for ( i=0; i < column_cnt; i++ )
	{
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
		column_name = (char *) sqlite3_column_name( pstmt->stmt, i );
		column_name_len = strlen( column_name );
		self->response.offset += response_xml_printf( R_BUF, column_name, column_name_len );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
	}
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

	for ( i=0; i < rows; i++ )
	{
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
		for ( p=0; p < column_cnt; p++ )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
			if ( self->db.pstmt_tree[i][p] != NULL )
			{
				data_len = strlen( self->db.pstmt_tree[i][p] );
				if ( data_len + 100 > R_BUF_LIMIT )
				{
					/* chunked response start */
					result = response_flush( self, FALSE );
					if ( result != SQLITE_OK )
					{
						self->db.pstmt_tree_len = rows+1;
						return result;
					}
				}
				self->response.offset += response_xml_printf( R_BUF, self->db.pstmt_tree[i][p], data_len );
			}
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
		}
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
	}
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );

	self->db.pstmt_tree_len = rows+1;
	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
		return response( self );
}
#undef R_BUF
#undef R_BUF_LIMIT

void pstmt_internal_gc( struct thread_t *self )
{
	int	i;

	if ( self->db.pstmt_tree == NULL || self->db.pstmt_tree_len <= 0 )
		return;

	for ( i=self->db.pstmt_tree_len; i--; )
	{
		if ( self->db.pstmt_tree[i] != NULL )
			free( self->db.pstmt_tree[i] );
	}
	memset( self->db.pstmt_tree, 0, sizeof(char *) * self->db.pstmt_tree_len );
	self->db.pstmt_tree_len = 0;
	return;
}

struct pstmt_t *pstmt_internal_find( struct thread_t *self, unsigned long pstmt_id, int *idx )
{
	int	i;

	if ( self->db.pstmt_cnt <= 0 )
		return NULL;

	for ( i=self->db.pstmt_cnt; i--; )
	{
		if ( self->db.pstmt[i].id == pstmt_id )
		{
			if ( idx != NULL )
				(*idx) = i;
			return &(self->db.pstmt[i]);
		}
	}
	return NULL;
}

int pstmt_internal_release( struct thread_t *self, unsigned long pstmt_id )
{
	int		i, p;
	struct pstmt_t	*pool = NULL;

	if ( self->db.pstmt_cnt <= 0 )
		return SQLITE_ERROR_PSTMT_EMPTY;

	for ( i=self->db.pstmt_cnt; i--; )
	{
		if ( self->db.pstmt[i].id == pstmt_id )
		{
			if ( self->db.pstmt[i].sql != NULL )
				free( self->db.pstmt[i].sql );

			if ( self->db.pstmt[i].params_cnt > 0 )
			{
				for ( p=0; p < self->db.pstmt[i].params_cnt; p++ )
				{
					if ( self->db.pstmt[i].params[p].type == 'T' && self->db.pstmt[i].params[p].data.value_text != NULL )
						free( self->db.pstmt[i].params[p].data.value_text );
					else if ( self->db.pstmt[i].params[p].type == 'B' && self->db.pstmt[i].params[p].data.value_blob != NULL )
						free( self->db.pstmt[i].params[p].data.value_blob );
				}
				free( self->db.pstmt[i].params );
			}

			sqlite3_finalize( self->db.pstmt[i].stmt );
			if ( self->db.pstmt_cnt > 1 )
			{
				pool = malloc( sizeof(struct pstmt_t) * (self->db.pstmt_cnt-1) );
				if ( pool == NULL )
					return SQLITE_ERROR;
				if ( i > 0 )
					memcpy( pool, self->db.pstmt, sizeof(struct pstmt_t) * i );
				if ( i < self->db.pstmt_cnt-1 )
					memcpy( pool, &(self->db.pstmt[i+1]), sizeof(struct pstmt_t) * (self->db.pstmt_cnt-(i+1)) );
			}
			else
				pool = NULL;
			free( self->db.pstmt );
			self->db.pstmt = pool;
			self->db.pstmt_cnt--;
			return SQLITE_OK;
		}
	}
	return SQLITE_NOTFOUND;
}

void pstmt_release_all( struct thread_t *self )
{
	while ( self->db.pstmt_cnt > 0 )
		pstmt_internal_release( self, self->db.pstmt[0].id );
}
